{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Lock Routines"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "This section is about the use of lock routines for synchronization."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### __omp_init_lock__ Routine"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The following example demonstrates how to initialize an array of locks in a __parallel__  region by using __omp_init_lock__."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "//%compiler: clang\n",
    "//%cflags: -fopenmp\n",
    "\n",
    "/*\n",
    "* name: init_lock.1\n",
    "* type: C++\n",
    "*/\n",
    "#include <omp.h>\n",
    "\n",
    "omp_lock_t *new_locks() {\n",
    "  int i;\n",
    "  omp_lock_t *lock = new omp_lock_t[1000];\n",
    "\n",
    "  #pragma omp parallel for private(i)\n",
    "    for (i=0; i<1000; i++)\n",
    "    { omp_init_lock(&lock[i]); }\n",
    "\n",
    "    return lock;\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!!%compiler: gfortran\n",
    "!!%cflags: -fopenmp\n",
    "\n",
    "! name: init_lock.1\n",
    "! type: F-fixed\n",
    "      FUNCTION NEW_LOCKS()\n",
    "        USE OMP_LIB        ! or INCLUDE \"omp_lib.h\"\n",
    "        INTEGER(OMP_LOCK_KIND), DIMENSION(1000) :: NEW_LOCKS\n",
    "        INTEGER I\n",
    "\n",
    "!$OMP   PARALLEL DO PRIVATE(I)\n",
    "          DO I=1,1000\n",
    "            CALL OMP_INIT_LOCK(NEW_LOCKS(I))\n",
    "          END DO\n",
    "!$OMP   END PARALLEL DO\n",
    "\n",
    "      END FUNCTION NEW_LOCKS"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### __omp_init_lock_with_hint__ Routine"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The following example demonstrates how to initialize an array of locks in a __parallel__ region by using __omp_init_lock_with_hint__. Note, hints are combined with an __|__ or __+__ operator in C/C++ and a __+__ operator in Fortran."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "//%compiler: clang\n",
    "//%cflags: -fopenmp\n",
    "\n",
    "/*\n",
    "* name: init_lock_with_hint.1\n",
    "* type: C++\n",
    "* version: omp_5.0\n",
    "*/\n",
    "#include <omp.h>\n",
    "\n",
    "omp_lock_t *new_locks()\n",
    "{\n",
    "  int i;\n",
    "  omp_lock_t *lock = new omp_lock_t[1000];\n",
    "\n",
    "  #pragma omp parallel for private(i)\n",
    "    for (i=0; i<1000; i++)\n",
    "    {\n",
    "      omp_init_lock_with_hint(&lock[i],\n",
    "        static_cast<omp_lock_hint_t>(omp_sync_hint_contended |\n",
    "                                     omp_sync_hint_speculative));\n",
    "    }\n",
    "    return lock;\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!!%compiler: gfortran\n",
    "!!%cflags: -fopenmp\n",
    "\n",
    "! name: init_lock_with_hint.1\n",
    "! type: F-fixed\n",
    "! version: omp_5.0\n",
    "      FUNCTION NEW_LOCKS()\n",
    "        USE OMP_LIB        ! or INCLUDE \"omp_lib.h\"\n",
    "        INTEGER(OMP_LOCK_KIND), DIMENSION(1000) :: NEW_LOCKS\n",
    "\n",
    "        INTEGER I\n",
    "\n",
    "!$OMP   PARALLEL DO PRIVATE(I)\n",
    "          DO I=1,1000\n",
    "            CALL OMP_INIT_LOCK_WITH_HINT(NEW_LOCKS(I),\n",
    "     &             OMP_SYNC_HINT_CONTENDED + OMP_SYNC_HINT_SPECULATIVE)\n",
    "          END DO\n",
    "!$OMP   END PARALLEL DO\n",
    "\n",
    "      END FUNCTION NEW_LOCKS"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Ownership of Locks"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Ownership of locks has changed since OpenMP 2.5. In OpenMP 2.5, locks are owned  by threads; so a lock released by the __omp_unset_lock__ routine must be  owned by the same thread executing the routine.  Beginning with OpenMP 3.0, locks are owned  by task regions; so a lock released by the __omp_unset_lock__ routine in  a task region must be owned by the same task region."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "This change in ownership requires extra care when using locks. The following program  is conforming in OpenMP 2.5 because the thread that releases the lock __lck__  in the parallel region is the same thread that acquired the lock in the sequential  part of the program (primary thread of parallel region and the initial thread are  the same). However, it is not conforming beginning with OpenMP 3.0, because the task  region that releases the lock __lck__ is different from the task region that  acquires the lock."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "//%compiler: clang\n",
    "//%cflags: -fopenmp\n",
    "\n",
    "/*\n",
    "* name: lock_owner.1\n",
    "* type: C\n",
    "* version:    omp_5.1\n",
    "*/\n",
    "#include <stdlib.h>\n",
    "#include <stdio.h>\n",
    "#include <omp.h>\n",
    "\n",
    "int main()\n",
    "{\n",
    "  int x;\n",
    "  omp_lock_t lck;\n",
    "\n",
    "  omp_init_lock (&lck);\n",
    "  omp_set_lock (&lck);\n",
    "  x = 0;\n",
    "\n",
    "#pragma omp parallel shared (x)\n",
    "  {\n",
    "    #pragma omp masked\n",
    "      {\n",
    "        x = x + 1;\n",
    "        omp_unset_lock (&lck);\n",
    "      }\n",
    "\n",
    "    /* Some more stuff. */\n",
    "  }\n",
    "  omp_destroy_lock (&lck);\n",
    "  return 0;\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!!%compiler: gfortran\n",
    "!!%cflags: -fopenmp\n",
    "\n",
    "! name: lock_owner.1\n",
    "! type: F-fixed\n",
    "! version:    omp_5.1\n",
    "        program lock\n",
    "        use omp_lib\n",
    "        integer :: x\n",
    "        integer (kind=omp_lock_kind) :: lck\n",
    "\n",
    "        call omp_init_lock (lck)\n",
    "        call omp_set_lock(lck)\n",
    "        x = 0\n",
    "\n",
    "!$omp parallel shared (x)\n",
    "!$omp masked\n",
    "        x = x + 1\n",
    "        call omp_unset_lock(lck)\n",
    "!$omp end masked\n",
    "\n",
    "!       Some more stuff.\n",
    "!$omp end parallel\n",
    "\n",
    "        call omp_destroy_lock(lck)\n",
    "\n",
    "        end"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Simple Lock Routines"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "In the following example, the lock routines cause the threads to be idle while  waiting for entry to the first critical section, but to do other work while waiting  for entry to the second. The __omp_set_lock__ function blocks, but the __omp_test_lock__  function does not, allowing the work in __skip__ to be done."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Note that the argument to the lock routines should have type  __omp_lock_t__ (or __omp_lock_kind__ in Fortran),  and that there is no need to flush the lock variable ( _lck_ )."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "//%compiler: clang\n",
    "//%cflags: -fopenmp\n",
    "\n",
    "/*\n",
    "* name: simple_lock.1\n",
    "* type: C\n",
    "*/\n",
    "#include <stdio.h>\n",
    "#include <omp.h>\n",
    "void skip(int i) {}\n",
    "void work(int i) {}\n",
    "int main()\n",
    "{\n",
    "  omp_lock_t lck;\n",
    "  int id;\n",
    "  omp_init_lock(&lck);\n",
    "\n",
    "  #pragma omp parallel shared(lck) private(id)\n",
    "  {\n",
    "    id = omp_get_thread_num();\n",
    "\n",
    "    omp_set_lock(&lck);\n",
    "    /*  only one thread at a time can execute this printf */\n",
    "    printf(\"My thread id is %d.\\n\", id);\n",
    "    omp_unset_lock(&lck);\n",
    "\n",
    "    while (! omp_test_lock(&lck)) {\n",
    "      skip(id);   /* we do not yet have the lock,\n",
    "                     so we must do something else */\n",
    "    }\n",
    "\n",
    "    work(id);      /* we now have the lock\n",
    "                      and can do the work */\n",
    "\n",
    "    omp_unset_lock(&lck);\n",
    "  }\n",
    "  omp_destroy_lock(&lck);\n",
    "\n",
    "  return 0;\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!!%compiler: gfortran\n",
    "!!%cflags: -fopenmp\n",
    "\n",
    "! name: simple_lock.1\n",
    "! type: F-fixed\n",
    "      SUBROUTINE SKIP(ID)\n",
    "      END SUBROUTINE SKIP\n",
    "\n",
    "      SUBROUTINE WORK(ID)\n",
    "      END SUBROUTINE WORK\n",
    "\n",
    "      PROGRAM SIMPLELOCK\n",
    "\n",
    "        INCLUDE \"omp_lib.h\"     ! or USE OMP_LIB\n",
    "\n",
    "        INTEGER(OMP_LOCK_KIND) LCK\n",
    "        INTEGER ID\n",
    "\n",
    "        CALL OMP_INIT_LOCK(LCK)\n",
    "\n",
    "!$OMP   PARALLEL SHARED(LCK) PRIVATE(ID)\n",
    "          ID = OMP_GET_THREAD_NUM()\n",
    "          CALL OMP_SET_LOCK(LCK)\n",
    "          PRINT *, 'My thread id is ', ID\n",
    "          CALL OMP_UNSET_LOCK(LCK)\n",
    "\n",
    "          DO WHILE (.NOT. OMP_TEST_LOCK(LCK))\n",
    "            CALL SKIP(ID)     ! We do not yet have the lock\n",
    "                              ! so we must do something else\n",
    "          END DO\n",
    "\n",
    "          CALL WORK(ID)       ! We now have the lock\n",
    "                              ! and can do the work\n",
    "\n",
    "          CALL OMP_UNSET_LOCK( LCK )\n",
    "\n",
    "!$OMP   END PARALLEL\n",
    "\n",
    "        CALL OMP_DESTROY_LOCK( LCK )\n",
    "\n",
    "      END PROGRAM SIMPLELOCK"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Nestable Lock Routines"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The following example demonstrates how a nestable lock can be used to synchronize  updates both to a whole structure and to one of its members."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "//%compiler: clang\n",
    "//%cflags: -fopenmp\n",
    "\n",
    "/*\n",
    "* name: nestable_lock.1\n",
    "* type: C\n",
    "*/\n",
    "\n",
    "#include <omp.h>\n",
    "\n",
    "typedef struct {\n",
    "   int a,b;\n",
    "   omp_nest_lock_t lck;\n",
    "} pair;\n",
    "\n",
    "int work1();\n",
    "int work2();\n",
    "int work3();\n",
    "\n",
    "void incr_a(pair *p, int a)\n",
    "{\n",
    "\n",
    "  /* Called only from incr_pair, no need to lock. */\n",
    "  p->a += a;\n",
    "\n",
    "}\n",
    "\n",
    "void incr_b(pair *p, int b)\n",
    "{\n",
    "\n",
    "  /* Called both from incr_pair and elsewhere, */\n",
    "  /* so need a nestable lock. */\n",
    "\n",
    "  omp_set_nest_lock(&p->lck);\n",
    "  p->b += b;\n",
    "  omp_unset_nest_lock(&p->lck);\n",
    "\n",
    "}\n",
    "\n",
    "void incr_pair(pair *p, int a, int b)\n",
    "{\n",
    "\n",
    "  omp_set_nest_lock(&p->lck);\n",
    "  incr_a(p, a);\n",
    "  incr_b(p, b);\n",
    "  omp_unset_nest_lock(&p->lck);\n",
    "\n",
    "}\n",
    "\n",
    "void nestlock(pair *p)\n",
    "{\n",
    "\n",
    "  #pragma omp parallel sections\n",
    "  {\n",
    "    #pragma omp section\n",
    "      incr_pair(p, work1(), work2());\n",
    "    #pragma omp section\n",
    "      incr_b(p, work3());\n",
    "  }\n",
    "\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!!%compiler: gfortran\n",
    "!!%cflags: -fopenmp\n",
    "\n",
    "! name: nestable_lock.1\n",
    "! type: F-fixed\n",
    "      MODULE DATA\n",
    "        USE OMP_LIB, ONLY: OMP_NEST_LOCK_KIND\n",
    "        TYPE LOCKED_PAIR\n",
    "          INTEGER A\n",
    "          INTEGER B\n",
    "          INTEGER (OMP_NEST_LOCK_KIND) LCK\n",
    "       END TYPE\n",
    "      END MODULE DATA\n",
    "\n",
    "      SUBROUTINE INCR_A(P, A)\n",
    "        ! called only from INCR_PAIR, no need to lock\n",
    "        USE DATA\n",
    "        TYPE(LOCKED_PAIR) :: P\n",
    "        INTEGER A\n",
    "        P%A = P%A + A\n",
    "      END SUBROUTINE INCR_A\n",
    "\n",
    "      SUBROUTINE INCR_B(P, B)\n",
    "        ! called from both INCR_PAIR and elsewhere,\n",
    "        ! so we need a nestable lock\n",
    "        USE OMP_LIB       ! or INCLUDE \"omp_lib.h\"\n",
    "        USE DATA\n",
    "        TYPE(LOCKED_PAIR) :: P\n",
    "        INTEGER B\n",
    "        CALL OMP_SET_NEST_LOCK(P%LCK)\n",
    "        P%B = P%B + B\n",
    "        CALL OMP_UNSET_NEST_LOCK(P%LCK)\n",
    "      END SUBROUTINE INCR_B\n",
    "\n",
    "      SUBROUTINE INCR_PAIR(P, A, B)\n",
    "        USE OMP_LIB        ! or INCLUDE \"omp_lib.h\"\n",
    "        USE DATA\n",
    "        TYPE(LOCKED_PAIR) :: P\n",
    "        INTEGER A\n",
    "        INTEGER B\n",
    "\n",
    "        CALL OMP_SET_NEST_LOCK(P%LCK)\n",
    "        CALL INCR_A(P, A)\n",
    "        CALL INCR_B(P, B)\n",
    "        CALL OMP_UNSET_NEST_LOCK(P%LCK)\n",
    "      END SUBROUTINE INCR_PAIR\n",
    "\n",
    "      SUBROUTINE NESTLOCK(P)\n",
    "        USE OMP_LIB        ! or INCLUDE \"omp_lib.h\"\n",
    "        USE DATA\n",
    "        TYPE(LOCKED_PAIR) :: P\n",
    "        INTEGER WORK1, WORK2, WORK3\n",
    "        EXTERNAL WORK1, WORK2, WORK3\n",
    "\n",
    "!$OMP   PARALLEL SECTIONS\n",
    "\n",
    "!$OMP   SECTION\n",
    "          CALL INCR_PAIR(P, WORK1(), WORK2())\n",
    "!$OMP   SECTION\n",
    "          CALL INCR_B(P, WORK3())\n",
    "!$OMP   END PARALLEL SECTIONS\n",
    "\n",
    "      END SUBROUTINE NESTLOCK"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Native",
   "language": "native",
   "name": "native"
  },
  "language_info": {
   "file_extension": ".c",
   "mimetype": "text/plain",
   "name": "c"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
